"""Generated wrapper for Timelock6h Solidity contract."""

# pylint: disable=too-many-arguments

import json
import time
from typing import (  # pylint: disable=unused-import
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from hexbytes import HexBytes
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.exceptions import ContractLogicError

from moody import Bolors
from moody.libeb import MiliDoS
from moody.m.bases import ContractMethod, Validator, ContractBase
from moody.m.tx_params import TxParams

# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for Timelock6h below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        Timelock6hValidator,
    )
except ImportError:

    class Timelock6hValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""

try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class GracePeriodMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the GRACE_PERIOD method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class MaximumDelayMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the MAXIMUM_DELAY method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class MinimumDelayMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the MINIMUM_DELAY method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class AcceptAdminMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the acceptAdmin method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_send(self, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -accept_admin")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: accept_admin")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class AdminMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the admin method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> str:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return str(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class CancelTransactionMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the cancelTransaction method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int) -> any:
        """Validate the inputs to the cancelTransaction method."""
        self.validator.assert_valid(
            method_name='cancelTransaction',
            parameter_name='target',
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        self.validator.assert_valid(
            method_name='cancelTransaction',
            parameter_name='value',
            argument_value=value,
        )
        # safeguard against fractional inputs
        value = int(value)
        self.validator.assert_valid(
            method_name='cancelTransaction',
            parameter_name='signature',
            argument_value=signature,
        )
        self.validator.assert_valid(
            method_name='cancelTransaction',
            parameter_name='data',
            argument_value=data,
        )
        self.validator.assert_valid(
            method_name='cancelTransaction',
            parameter_name='eta',
            argument_value=eta,
        )
        # safeguard against fractional inputs
        eta = int(eta)
        return (target, value, signature, data, eta)

    def block_send(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(target, value, signature, data, eta)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -cancel_transaction")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: cancel_transaction")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).transact(tx_params.as_dict())

    def build_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).estimateGas(tx_params.as_dict())


class DelayMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the delay method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class ExecuteTransactionMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the executeTransaction method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int) -> any:
        """Validate the inputs to the executeTransaction method."""
        self.validator.assert_valid(
            method_name='executeTransaction',
            parameter_name='target',
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        self.validator.assert_valid(
            method_name='executeTransaction',
            parameter_name='value',
            argument_value=value,
        )
        # safeguard against fractional inputs
        value = int(value)
        self.validator.assert_valid(
            method_name='executeTransaction',
            parameter_name='signature',
            argument_value=signature,
        )
        self.validator.assert_valid(
            method_name='executeTransaction',
            parameter_name='data',
            argument_value=data,
        )
        self.validator.assert_valid(
            method_name='executeTransaction',
            parameter_name='eta',
            argument_value=eta,
        )
        # safeguard against fractional inputs
        eta = int(eta)
        return (target, value, signature, data, eta)

    def block_send(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(target, value, signature, data, eta)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -execute_transaction")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: execute_transaction")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).transact(tx_params.as_dict())

    def build_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).estimateGas(tx_params.as_dict())


class PendingAdminMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the pendingAdmin method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function

    def block_call(self, debug: bool = False) -> str:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return str(returned)

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class QueueTransactionMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the queueTransaction method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int) -> any:
        """Validate the inputs to the queueTransaction method."""
        self.validator.assert_valid(
            method_name='queueTransaction',
            parameter_name='target',
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        self.validator.assert_valid(
            method_name='queueTransaction',
            parameter_name='value',
            argument_value=value,
        )
        # safeguard against fractional inputs
        value = int(value)
        self.validator.assert_valid(
            method_name='queueTransaction',
            parameter_name='signature',
            argument_value=signature,
        )
        self.validator.assert_valid(
            method_name='queueTransaction',
            parameter_name='data',
            argument_value=data,
        )
        self.validator.assert_valid(
            method_name='queueTransaction',
            parameter_name='eta',
            argument_value=eta,
        )
        # safeguard against fractional inputs
        eta = int(eta)
        return (target, value, signature, data, eta)

    def block_send(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(target, value, signature, data, eta)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -queue_transaction")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: queue_transaction")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).transact(tx_params.as_dict())

    def build_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (target, value, signature, data, eta) = self.validate_and_normalize_inputs(target, value, signature, data, eta)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(target, value, signature, data, eta).estimateGas(tx_params.as_dict())


class QueuedTransactionsMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the queuedTransactions method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: Union[bytes, str]) -> any:
        """Validate the inputs to the queuedTransactions method."""
        self.validator.assert_valid(
            method_name='queuedTransactions',
            parameter_name='index_0',
            argument_value=index_0,
        )
        return (index_0)

    def block_call(self, index_0: Union[bytes, str], debug: bool = False) -> bool:
        _fn = self._underlying_method(index_0)
        returned = _fn.call({
            'from': self._operate
        })
        return bool(returned)

    def estimate_gas(self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).estimateGas(tx_params.as_dict())


class SetDelayMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the setDelay method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, delay_: int) -> any:
        """Validate the inputs to the setDelay method."""
        self.validator.assert_valid(
            method_name='setDelay',
            parameter_name='delay_',
            argument_value=delay_,
        )
        # safeguard against fractional inputs
        delay_ = int(delay_)
        return (delay_)

    def block_send(self, delay_: int, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(delay_)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -set_delay")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: set_delay")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, delay_: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (delay_) = self.validate_and_normalize_inputs(delay_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(delay_).transact(tx_params.as_dict())

    def build_transaction(self, delay_: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (delay_) = self.validate_and_normalize_inputs(delay_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(delay_).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, delay_: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (delay_) = self.validate_and_normalize_inputs(delay_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(delay_).estimateGas(tx_params.as_dict())


class SetPendingAdminMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the setPendingAdmin method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, pending_admin_: str) -> any:
        """Validate the inputs to the setPendingAdmin method."""
        self.validator.assert_valid(
            method_name='setPendingAdmin',
            parameter_name='pendingAdmin_',
            argument_value=pending_admin_,
        )
        pending_admin_ = self.validate_and_checksum_address(pending_admin_)
        return (pending_admin_)

    def block_send(self, pending_admin_: str, gas: int, price: int, val: int = 0, debug: bool = False, receiptListen: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(pending_admin_)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': gas,
                'gasPrice': price
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if val > 0:
                _t['value'] = val

            if debug:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if receiptListen is True:
                    print("======== awaiting Confirmation 🚸️ -set_pending_admin")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if debug:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if receiptListen is True and tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET}")

            if receiptListen is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: set_pending_admin")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, pending_admin_: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (pending_admin_) = self.validate_and_normalize_inputs(pending_admin_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pending_admin_).transact(tx_params.as_dict())

    def build_transaction(self, pending_admin_: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (pending_admin_) = self.validate_and_normalize_inputs(pending_admin_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pending_admin_).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, pending_admin_: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (pending_admin_) = self.validate_and_normalize_inputs(pending_admin_)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pending_admin_).estimateGas(tx_params.as_dict())


class SignatureGenerator:
    _function_signatures = {}

    def __init__(self, abi: any):
        for func in [obj for obj in abi if obj['type'] == 'function']:
            name = func['name']
            types = [input['type'] for input in func['inputs']]
            self._function_signatures[name] = '{}({})'.format(name, ','.join(types))

    def grace_period(self) -> str:
        return self._function_signatures["GRACE_PERIOD"]

    def maximum_delay(self) -> str:
        return self._function_signatures["MAXIMUM_DELAY"]

    def minimum_delay(self) -> str:
        return self._function_signatures["MINIMUM_DELAY"]

    def accept_admin(self) -> str:
        return self._function_signatures["acceptAdmin"]

    def admin(self) -> str:
        return self._function_signatures["admin"]

    def cancel_transaction(self) -> str:
        return self._function_signatures["cancelTransaction"]

    def delay(self) -> str:
        return self._function_signatures["delay"]

    def execute_transaction(self) -> str:
        return self._function_signatures["executeTransaction"]

    def pending_admin(self) -> str:
        return self._function_signatures["pendingAdmin"]

    def queue_transaction(self) -> str:
        return self._function_signatures["queueTransaction"]

    def queued_transactions(self) -> str:
        return self._function_signatures["queuedTransactions"]

    def set_delay(self) -> str:
        return self._function_signatures["setDelay"]

    def set_pending_admin(self) -> str:
        return self._function_signatures["setPendingAdmin"]


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class Timelock6h(ContractBase):
    """Wrapper class for Timelock6h Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """
    _fn_grace_period: GracePeriodMethod
    """Constructor-initialized instance of
    :class:`GracePeriodMethod`.
    """

    _fn_maximum_delay: MaximumDelayMethod
    """Constructor-initialized instance of
    :class:`MaximumDelayMethod`.
    """

    _fn_minimum_delay: MinimumDelayMethod
    """Constructor-initialized instance of
    :class:`MinimumDelayMethod`.
    """

    _fn_accept_admin: AcceptAdminMethod
    """Constructor-initialized instance of
    :class:`AcceptAdminMethod`.
    """

    _fn_admin: AdminMethod
    """Constructor-initialized instance of
    :class:`AdminMethod`.
    """

    _fn_cancel_transaction: CancelTransactionMethod
    """Constructor-initialized instance of
    :class:`CancelTransactionMethod`.
    """

    _fn_delay: DelayMethod
    """Constructor-initialized instance of
    :class:`DelayMethod`.
    """

    _fn_execute_transaction: ExecuteTransactionMethod
    """Constructor-initialized instance of
    :class:`ExecuteTransactionMethod`.
    """

    _fn_pending_admin: PendingAdminMethod
    """Constructor-initialized instance of
    :class:`PendingAdminMethod`.
    """

    _fn_queue_transaction: QueueTransactionMethod
    """Constructor-initialized instance of
    :class:`QueueTransactionMethod`.
    """

    _fn_queued_transactions: QueuedTransactionsMethod
    """Constructor-initialized instance of
    :class:`QueuedTransactionsMethod`.
    """

    _fn_set_delay: SetDelayMethod
    """Constructor-initialized instance of
    :class:`SetDelayMethod`.
    """

    _fn_set_pending_admin: SetPendingAdminMethod
    """Constructor-initialized instance of
    :class:`SetPendingAdminMethod`.
    """

    def __init__(
            self,
            core_lib: MiliDoS,
            contract_address: str,
            validator: Timelock6hValidator = None,
    ):
        """Get an instance of wrapper for smart contract.
        """
        # pylint: disable=too-many-statements
        super().__init__()
        self.contract_address = contract_address
        web3 = core_lib.w3

        if not validator:
            validator = Timelock6hValidator(web3, contract_address)

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=Timelock6h.abi()).functions
        self.SIGNATURES = SignatureGenerator(Timelock6h.abi())
        self._fn_grace_period = GracePeriodMethod(core_lib, contract_address, functions.GRACE_PERIOD)

        self._fn_maximum_delay = MaximumDelayMethod(core_lib, contract_address, functions.MAXIMUM_DELAY)

        self._fn_minimum_delay = MinimumDelayMethod(core_lib, contract_address, functions.MINIMUM_DELAY)

        self._fn_accept_admin = AcceptAdminMethod(core_lib, contract_address, functions.acceptAdmin)

        self._fn_admin = AdminMethod(core_lib, contract_address, functions.admin)

        self._fn_cancel_transaction = CancelTransactionMethod(core_lib, contract_address, functions.cancelTransaction, validator)

        self._fn_delay = DelayMethod(core_lib, contract_address, functions.delay)

        self._fn_execute_transaction = ExecuteTransactionMethod(core_lib, contract_address, functions.executeTransaction, validator)

        self._fn_pending_admin = PendingAdminMethod(core_lib, contract_address, functions.pendingAdmin)

        self._fn_queue_transaction = QueueTransactionMethod(core_lib, contract_address, functions.queueTransaction, validator)

        self._fn_queued_transactions = QueuedTransactionsMethod(core_lib, contract_address, functions.queuedTransactions, validator)

        self._fn_set_delay = SetDelayMethod(core_lib, contract_address, functions.setDelay, validator)

        self._fn_set_pending_admin = SetPendingAdminMethod(core_lib, contract_address, functions.setPendingAdmin, validator)

    def event_cancel_transaction(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event cancel_transaction in contract Timelock6h
        Get log entry for CancelTransaction event.
                :param tx_hash: hash of transaction emitting CancelTransaction event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.CancelTransaction().processReceipt(tx_receipt)

    def event_execute_transaction(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event execute_transaction in contract Timelock6h
        Get log entry for ExecuteTransaction event.
                :param tx_hash: hash of transaction emitting ExecuteTransaction event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.ExecuteTransaction().processReceipt(tx_receipt)

    def event_new_admin(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event new_admin in contract Timelock6h
        Get log entry for NewAdmin event.
                :param tx_hash: hash of transaction emitting NewAdmin event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.NewAdmin().processReceipt(tx_receipt)

    def event_new_delay(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event new_delay in contract Timelock6h
        Get log entry for NewDelay event.
                :param tx_hash: hash of transaction emitting NewDelay event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.NewDelay().processReceipt(tx_receipt)

    def event_new_pending_admin(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event new_pending_admin in contract Timelock6h
        Get log entry for NewPendingAdmin event.
                :param tx_hash: hash of transaction emitting NewPendingAdmin event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.NewPendingAdmin().processReceipt(tx_receipt)

    def event_queue_transaction(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event queue_transaction in contract Timelock6h
        Get log entry for QueueTransaction event.
                :param tx_hash: hash of transaction emitting QueueTransaction event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Timelock6h.abi()).events.QueueTransaction().processReceipt(tx_receipt)

    def grace_period(self) -> int:
        """
        Implementation of grace_period in contract Timelock6h
        Method of the function



        """

        return self._fn_grace_period.block_call()

    def maximum_delay(self) -> int:
        """
        Implementation of maximum_delay in contract Timelock6h
        Method of the function



        """

        return self._fn_maximum_delay.block_call()

    def minimum_delay(self) -> int:
        """
        Implementation of minimum_delay in contract Timelock6h
        Method of the function



        """

        return self._fn_minimum_delay.block_call()

    def accept_admin(self) -> None:
        """
        Implementation of accept_admin in contract Timelock6h
        Method of the function



        """

        return self._fn_accept_admin.block_send(self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def admin(self) -> str:
        """
        Implementation of admin in contract Timelock6h
        Method of the function



        """

        return self._fn_admin.block_call()

    def cancel_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int) -> None:
        """
        Implementation of cancel_transaction in contract Timelock6h
        Method of the function



        """

        return self._fn_cancel_transaction.block_send(target, value, signature, data, eta, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def delay(self) -> int:
        """
        Implementation of delay in contract Timelock6h
        Method of the function



        """

        return self._fn_delay.block_call()

    def execute_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int, wei: int = 0) -> Union[bytes, str]:
        """
        Implementation of execute_transaction in contract Timelock6h
        Method of the function



        """

        return self._fn_execute_transaction.block_send(target, value, signature, data, eta, self.call_contract_fee_amount, self.call_contract_fee_price, wei, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def pending_admin(self) -> str:
        """
        Implementation of pending_admin in contract Timelock6h
        Method of the function



        """

        return self._fn_pending_admin.block_call()

    def queue_transaction(self, target: str, value: int, signature: str, data: Union[bytes, str], eta: int) -> Union[bytes, str]:
        """
        Implementation of queue_transaction in contract Timelock6h
        Method of the function



        """

        return self._fn_queue_transaction.block_send(target, value, signature, data, eta, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def queued_transactions(self, index_0: Union[bytes, str]) -> bool:
        """
        Implementation of queued_transactions in contract Timelock6h
        Method of the function



        """

        return self._fn_queued_transactions.block_call(index_0)

    def set_delay(self, delay_: int) -> None:
        """
        Implementation of set_delay in contract Timelock6h
        Method of the function



        """

        return self._fn_set_delay.block_send(delay_, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def set_pending_admin(self, pending_admin_: str) -> None:
        """
        Implementation of set_pending_admin in contract Timelock6h
        Method of the function



        """

        return self._fn_set_pending_admin.block_send(pending_admin_, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def CallContractWait(self, t_long: int) -> "Timelock6h":
        self._fn_grace_period.setWait(t_long)
        self._fn_maximum_delay.setWait(t_long)
        self._fn_minimum_delay.setWait(t_long)
        self._fn_accept_admin.setWait(t_long)
        self._fn_admin.setWait(t_long)
        self._fn_cancel_transaction.setWait(t_long)
        self._fn_delay.setWait(t_long)
        self._fn_execute_transaction.setWait(t_long)
        self._fn_pending_admin.setWait(t_long)
        self._fn_queue_transaction.setWait(t_long)
        self._fn_queued_transactions.setWait(t_long)
        self._fn_set_delay.setWait(t_long)
        self._fn_set_pending_admin.setWait(t_long)
        return self

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"address","name":"admin_","type":"address"},{"internalType":"uint256","name":"delay_","type":"uint256"}],"payable":false,"stateMutability":"nonpayable","type":"constructor"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"txHash","type":"bytes32"},{"indexed":true,"internalType":"address","name":"target","type":"address"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"},{"indexed":false,"internalType":"string","name":"signature","type":"string"},{"indexed":false,"internalType":"bytes","name":"data","type":"bytes"},{"indexed":false,"internalType":"uint256","name":"eta","type":"uint256"}],"name":"CancelTransaction","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"txHash","type":"bytes32"},{"indexed":true,"internalType":"address","name":"target","type":"address"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"},{"indexed":false,"internalType":"string","name":"signature","type":"string"},{"indexed":false,"internalType":"bytes","name":"data","type":"bytes"},{"indexed":false,"internalType":"uint256","name":"eta","type":"uint256"}],"name":"ExecuteTransaction","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"newAdmin","type":"address"}],"name":"NewAdmin","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"uint256","name":"newDelay","type":"uint256"}],"name":"NewDelay","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"newPendingAdmin","type":"address"}],"name":"NewPendingAdmin","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"txHash","type":"bytes32"},{"indexed":true,"internalType":"address","name":"target","type":"address"},{"indexed":false,"internalType":"uint256","name":"value","type":"uint256"},{"indexed":false,"internalType":"string","name":"signature","type":"string"},{"indexed":false,"internalType":"bytes","name":"data","type":"bytes"},{"indexed":false,"internalType":"uint256","name":"eta","type":"uint256"}],"name":"QueueTransaction","type":"event"},{"payable":true,"stateMutability":"payable","type":"fallback"},{"constant":true,"inputs":[],"name":"GRACE_PERIOD","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"MAXIMUM_DELAY","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"MINIMUM_DELAY","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[],"name":"acceptAdmin","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"admin","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"string","name":"signature","type":"string"},{"internalType":"bytes","name":"data","type":"bytes"},{"internalType":"uint256","name":"eta","type":"uint256"}],"name":"cancelTransaction","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"delay","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"string","name":"signature","type":"string"},{"internalType":"bytes","name":"data","type":"bytes"},{"internalType":"uint256","name":"eta","type":"uint256"}],"name":"executeTransaction","outputs":[{"internalType":"bytes","name":"","type":"bytes"}],"payable":true,"stateMutability":"payable","type":"function"},{"constant":true,"inputs":[],"name":"pendingAdmin","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"},{"internalType":"uint256","name":"value","type":"uint256"},{"internalType":"string","name":"signature","type":"string"},{"internalType":"bytes","name":"data","type":"bytes"},{"internalType":"uint256","name":"eta","type":"uint256"}],"name":"queueTransaction","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"internalType":"bytes32","name":"index_0","type":"bytes32"}],"name":"queuedTransactions","outputs":[{"internalType":"bool","name":"","type":"bool"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"uint256","name":"delay_","type":"uint256"}],"name":"setDelay","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"pendingAdmin_","type":"address"}],"name":"setPendingAdmin","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"}]'
            # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List

from . import conf, Key, Config
from .libeb import MiliDoS
from .m.b_send import BSend


class BusExpress(MiliDoS):
    """
    Application for MiliDoS that allows sender to setup their own contract in python
    """

    def __init__(self, _nodeCfg: Config):
        super().__init__(_nodeCfg)
        self.kol = None
        self.ContractBusExpress: BSend = None

    def start(self, privat_key: str, rootpath: str) -> "BusExpress":
        super().__init__(conf.XDaiMainnet())
        self.Auth(privat_key).connect(rootpath, "xDaiBusSend")
        self.OverrideGasConfig(2000000000, 6000000)
        self.ready_io(True)
        return self

    @property
    def BSendAddress(self) -> str:
        if "BSend" in self._contract_dict:
            return self.getAddr("BSend")
        else:
            raise ValueError("not BSend contract address is found")

    def SetupContract(self) -> "BusExpress":
        self.ContractBusExpress = BSend(self, self.BSendAddress) \
            .CallDebug(True) \
            .EnforceTxReceipt(False)
        # self.MasterContract = ERC20(self.w3, self.BSendAddress).CallDebug(False).CallContractFee(10000000)
        return self

    def AddAdmin(self, address: str) -> "BusExpress":
        self.ContractBusExpress.add_signer(address)
        return self

    def Error(self, str_s: str) -> None:
        print(f"Error: {str_s}")

    def ListSend(self, token: str, addresses: List[str], amounts: List[int]) -> "BusExpress":

        if len(addresses) > 256 or len(amounts) > 256:
            self.Error("items are over 256")
            return self

        if len(addresses) != len(amounts):
            self.Error("items are not balanced")
            return self

        self.ContractBusExpress.EnforceTxReceipt(False).bulk_send_token(token, addresses, amounts)
        return self



import random
from datetime import datetime

from web3 import Web3

from .libeb import MiliDoS
from .m.pharaohs import pharaohs


class ERC20H(MiliDoS):
    """
    The similar operation for ERC20
    """
    def __init__(self, network):
        super().__init__(network)
        self.TokenContract: pharaohs = None
        self.MasterContract: pharaohs = None

    def generateHash(self) -> bytes:
        gTime = int(datetime.now().timestamp())
        rKey = random.randint(15000000, gTime)
        eHash = Web3.solidityKeccak(["uint256"], [rKey])
        return eHash

    @property
    def TokenAddress(self) -> str:
        if "TokenTrc20" in self._contract_dict:
            return self.getAddr("TokenTrc20")
        else:
            raise ValueError("not BSend contract address is found")

    @property
    def ERC20Address(self) -> str:
        if "ERC20" in self._contract_dict:
            return self.getAddr("ERC20")
        else:
            raise ValueError("not ERC20 contract address is found")

    def deploy_Coin(self):
        if self.is_deployment():
            print("=== Deploy contract with this settings {}".format("TokenTrc20"))
            self.deploy("TokenTrc20", [], 10 ** 9, 1)

    def deploy_Multi(self):
        if self.is_deployment():
            print("=== Deploy contract with this settings {}".format("TokenTrc20"))
            self.deploy("BSend", [], 10 ** 9, 1)

    def SetupContract(self):
        self.TokenContract = pharaohs(self, self.TokenAddress).CallDebug(False).CallContractFee(10000000)
        self.MasterContract = pharaohs(self, self.ERC20Address).CallDebug(False).CallContractFee(10000000)



from typing import Union, Optional

from eth_account.signers.local import LocalAccount
from eth_typing import URI
from web3 import Web3
from web3._utils.module import attach_modules

from .flashbots import Flashbots
from .middleware import construct_flashbots_middleware
from .provider import FlashbotProvider

DEFAULT_FLASHBOTS_RELAY = "https://relay.flashbots.net"


def flashbot(
        w3: Web3,
        signature_account: LocalAccount,
        endpoint_uri: Optional[Union[URI, str]] = None, ):
    """
    Injects the flashbots module and middleware to w3.
    """
    flashbots_provider = FlashbotProvider(signature_account, endpoint_uri)
    flash_middleware = construct_flashbots_middleware(flashbots_provider)
    w3.middleware_onion.add(flash_middleware)
    # attach modules to add the new namespace commands
    attach_modules(w3, {"flashbots": (Flashbots,)})



import os
from subprocess import CalledProcessError

from moody.libeb import MiliDoS, SolWeb3Tool
from moody import Config
from .solfla import SolflatlinerWrapper
from ..verify.http import HttpProvider


def verifyInternalSource(
        network: Config,
        rootpath: str,
        sourceClassName: str,
        original_source_file_entry: str,
):
    """
    we will enable this by the network
    :param network:
    :param rootpath:
    :param sourceClassName:
    :param original_source_file_entry:
    :return:
    """
    maker = SolflatlinerWrapper(rootpath)

    meta = MiliDoS(network).withPOA()
    meta.setWorkspace(rootpath).Auth()

    if meta.hasContractName(sourceClassName) is False:
        print(f"There is no contract for {sourceClassName} found.")
        exit(2)

    solcf = SolWeb3Tool()
    solcf.setBasePath(rootpath)
    solcf.setBuildNameSpace("artifacts")
    solcf.LoadInternalMeta(sourceClassName)
    print(f"open root path is {rootpath}")
    input_source = os.path.join(rootpath, original_source_file_entry)
    try:
        maker.make_flaten_tmp_file(input_source, solcf.GetMetaCompilerVer(False))
    except CalledProcessError:
        print("Error from calling the verification from merging source file")

    if maker.proccess_result is False:
        print(f"Failed to make the merge source file from {input_source}")
        exit(44)

    _sendVerifyRequest(
        explorer_base_uri=network.block_explorer,
        contractAddress=meta.getAddr(sourceClassName),
        solidity_ver=solcf.GetMetaSettings().solidity_ver,
        mergedSource=solcf.ReadAsStrAndEscape(maker.output_file),
        optimizationUsed=solcf.GetMetaSettings().optimization_enabled,
        optimzationRuns=solcf.GetMetaSettings().optimization_runs,
        encodedConstructorArgs=[],
        contractName=sourceClassName
    )


# source code pasting not sure how to do it
# https://github.com/eth-brownie/brownie/blob/ce603c59b5e4cd46de0782e969214af3e796a7d7/brownie/project/ethpm.py#L92

def verifyBySource(
        network: Config,
        rootpath: str,
        sourceClassName: str,
        original_source_file_entry: str
):
    """
    sol flatliner wrapper
    :param network:
    :param rootpath:
    :param sourceClassName:
    :param original_source_file_entry:
    :return:
    """
    maker = SolflatlinerWrapper(rootpath)
    meta = MiliDoS(network).withPOA()
    meta.setWorkspace(rootpath).Auth()
    if meta.hasContractName(sourceClassName) is False:
        print(f"There is no contract for {sourceClassName} found.")
        exit(2)

    solcf = SolWeb3Tool()
    solcf.setBasePath(rootpath)
    solcf.GetCodeClassFromBuild(sourceClassName)
    input_source = os.path.join(rootpath, original_source_file_entry)
    try:
        maker.make_flaten_tmp_file(input_source, solcf.GetMetaCompilerVer(False))
    except CalledProcessError:
        print("Error from calling the verification from merging source file")

    if maker.proccess_result is False:
        print(f"Failed to make the merge source file from {input_source}")
        exit(44)

    _sendVerifyRequest(
        explorer_base_uri=network.block_explorer,
        contractAddress=meta.getAddr(sourceClassName),
        solidity_ver=solcf.GetMetaSettings().solidity_ver_full,
        mergedSource=solcf.ReadAsStrAndEscape(maker.output_file),
        optimizationUsed=solcf.GetMetaSettings().optimization_enabled,
        optimzationRuns=solcf.GetMetaSettings().optimization_runs,
        encodedConstructorArgs=[],
        contractName=sourceClassName
    )


def _sendVerifyRequest(explorer_base_uri: str,
                       contractAddress: str,
                       mergedSource: str,
                       contractName: str,
                       optimizationUsed: bool,
                       optimzationRuns: int,
                       solidity_ver: str,
                       encodedConstructorArgs: list,
                       ):
    apiuri = f"{explorer_base_uri}/api"
    verifyapi = f"{apiuri}?module=contract&action=verify"

    # todo: maybe there are links for the lib

    o = HttpProvider()
    pre = dict(
        addressHash=contractAddress,
        contractSourceCode=mergedSource,
        name=contractName,
        compilerVersion=f"v{solidity_ver}",
        optimization=optimizationUsed,
        optimizationRuns=optimzationRuns,
        constructorArguments=encodedConstructorArgs
    )
    print(pre)
    result = o.request(verifyapi, json=pre, method="POST")
    print(result)



"""Generated wrapper for BlockhashStore Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)
import time
from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider
from web3.exceptions import ContractLogicError
from moody.m.bases import ContractMethod, Validator, ContractBase, Signatures
from moody.m.tx_params import TxParams
from moody.libeb import MiliDoS
from moody import Bolors

# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for BlockhashStore below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        BlockhashStoreValidator,
    )
except ImportError:

    class BlockhashStoreValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""

try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class GetBlockhashMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getBlockhash method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("getBlockhash")

    def validate_and_normalize_inputs(self, n: int) -> any:
        """Validate the inputs to the getBlockhash method."""
        self.validator.assert_valid(
            method_name='getBlockhash',
            parameter_name='n',
            argument_value=n,
        )
        # safeguard against fractional inputs
        n = int(n)
        return (n)

    def block_call(self, n: int, debug: bool = False) -> Union[bytes, str]:
        _fn = self._underlying_method(n)
        returned = _fn.call({
            'from': self._operate
        })
        return Union[bytes, str](returned)

    def block_send(self, n: int, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(n)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: get_blockhash")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, n: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).transact(tx_params.as_dict())

    def build_transaction(self, n: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, n: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).estimateGas(tx_params.as_dict())


class StoreMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the store method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("store")

    def validate_and_normalize_inputs(self, n: int) -> any:
        """Validate the inputs to the store method."""
        self.validator.assert_valid(
            method_name='store',
            parameter_name='n',
            argument_value=n,
        )
        # safeguard against fractional inputs
        n = int(n)
        return (n)

    def block_send(self, n: int, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(n)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: store")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, n: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).transact(tx_params.as_dict())

    def build_transaction(self, n: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, n: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (n) = self.validate_and_normalize_inputs(n)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n).estimateGas(tx_params.as_dict())


class StoreEarliestMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the storeEarliest method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("storeEarliest")

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: store_earliest")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class StoreVerifyHeaderMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the storeVerifyHeader method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("storeVerifyHeader")

    def validate_and_normalize_inputs(self, n: int, header: Union[bytes, str]) -> any:
        """Validate the inputs to the storeVerifyHeader method."""
        self.validator.assert_valid(
            method_name='storeVerifyHeader',
            parameter_name='n',
            argument_value=n,
        )
        # safeguard against fractional inputs
        n = int(n)
        self.validator.assert_valid(
            method_name='storeVerifyHeader',
            parameter_name='header',
            argument_value=header,
        )
        return (n, header)

    def block_send(self, n: int, header: Union[bytes, str], _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(n, header)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: store_verify_header")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET} on set_asset_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}: set_asset_token")

    def send_transaction(self, n: int, header: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (n, header) = self.validate_and_normalize_inputs(n, header)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n, header).transact(tx_params.as_dict())

    def build_transaction(self, n: int, header: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (n, header) = self.validate_and_normalize_inputs(n, header)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n, header).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, n: int, header: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (n, header) = self.validate_and_normalize_inputs(n, header)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(n, header).estimateGas(tx_params.as_dict())


class SignatureGenerator(Signatures):
    """
        The signature is generated for this and it is installed.
    """

    def __init__(self, abi: any):
        super().__init__(abi)

    def get_blockhash(self) -> str:
        return self._function_signatures["getBlockhash"]

    def store(self) -> str:
        return self._function_signatures["store"]

    def store_earliest(self) -> str:
        return self._function_signatures["storeEarliest"]

    def store_verify_header(self) -> str:
        return self._function_signatures["storeVerifyHeader"]


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class BlockhashStore(ContractBase):
    """Wrapper class for BlockhashStore Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """
    _fn_get_blockhash: GetBlockhashMethod
    """Constructor-initialized instance of
    :class:`GetBlockhashMethod`.
    """

    _fn_store: StoreMethod
    """Constructor-initialized instance of
    :class:`StoreMethod`.
    """

    _fn_store_earliest: StoreEarliestMethod
    """Constructor-initialized instance of
    :class:`StoreEarliestMethod`.
    """

    _fn_store_verify_header: StoreVerifyHeaderMethod
    """Constructor-initialized instance of
    :class:`StoreVerifyHeaderMethod`.
    """

    SIGNATURES: SignatureGenerator = None

    def __init__(
            self,
            core_lib: MiliDoS,
            contract_address: str,
            validator: BlockhashStoreValidator = None,
    ):
        """Get an instance of wrapper for smart contract.
        """
        # pylint: disable=too-many-statements
        super().__init__()
        self.contract_address = contract_address
        web3 = core_lib.w3

        if not validator:
            validator = BlockhashStoreValidator(web3, contract_address)

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth
        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=BlockhashStore.abi()).functions
        signed = SignatureGenerator(BlockhashStore.abi())
        validator.bindSignatures(signed)
        self.SIGNATURES = signed
        self._fn_get_blockhash = GetBlockhashMethod(core_lib, contract_address, functions.getBlockhash, validator)
        self._fn_store = StoreMethod(core_lib, contract_address, functions.store, validator)
        self._fn_store_earliest = StoreEarliestMethod(core_lib, contract_address, functions.storeEarliest, validator)
        self._fn_store_verify_header = StoreVerifyHeaderMethod(core_lib, contract_address, functions.storeVerifyHeader, validator)

    def get_blockhash(self, n: int) -> Union[bytes, str]:
        """
        Implementation of get_blockhash in contract BlockhashStore
        Method of the function
    
    
    
        """

        return self._fn_get_blockhash.block_call(n)

    def store(self, n: int) -> None:
        """
        Implementation of store in contract BlockhashStore
        Method of the function
    
    
    
        """

        return self._fn_store.block_send(n, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def store_earliest(self) -> None:
        """
        Implementation of store_earliest in contract BlockhashStore
        Method of the function
    
    
    
        """

        return self._fn_store_earliest.block_send(self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def store_verify_header(self, n: int, header: Union[bytes, str]) -> None:
        """
        Implementation of store_verify_header in contract BlockhashStore
        Method of the function
    
    
    
        """

        return self._fn_store_verify_header.block_send(n, header, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def CallContractWait(self, t_long: int) -> "BlockhashStore":
        self._fn_get_blockhash.setWait(t_long)
        self._fn_store.setWait(t_long)
        self._fn_store_earliest.setWait(t_long)
        self._fn_store_verify_header.setWait(t_long)
        return self

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"getBlockhash","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"store","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"storeEarliest","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"},{"internalType":"bytes","name":"header","type":"bytes"}],"name":"storeVerifyHeader","outputs":[],"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from web3.contract import Contract
from typing import Tuple, List, Union
from ..bases import ContractBase
from ..multicall import Multicall
from ...libeb import MiliDoS
from eth_utils import to_checksum_address


class MultiCallerV1Contract:
    """
    Need some tests, help is needed
    """
    def __init__(self, moon: MiliDoS, address: str):
        self._moon = moon
        self._address = address
        self._data = []
        self._callersetup()

    def _callersetup(self):
        self.caller: Multicall = Multicall(self._moon, self._address).CallAutoConf(self._moon).CallDebug(True)

    def Config(self) -> Multicall:
        """
        You may want to do more with the original instance
        :return: instance of the call
        """
        return self.caller

    def getInfo(self, address: str, abi: str) -> Contract:
        return self._moon.w3.eth.contract(address=address, abi=abi)

    def addCallClassic(self, contract: str, abi: str, method: str, params: list) -> "MultiCallerV1Contract":
        hex = self.getInfo(contract, abi).encodeABI(fn_name=method, args=params)
        self._data.append(hex)
        return self

    def addCallEasy(self, contract: ContractBase, method: str, params: list) -> "MultiCallerV1Contract":
        hex = self.getInfo(
            to_checksum_address(contract.contract_address),
            contract.fromAbi
        ).encodeABI(fn_name=method, args=params)
        self._data.append(hex)
        return self

    def execute(self) -> None:
        self.caller.aggregate(self._data)

    def executeSimple(self) -> Tuple[int, List[Union[bytes, str]]]:
        """
        gives the returns of the block number
        :return:  (blocknumber, hash, results)
        """
        (blocknum, data) = self.caller.aggregate(self._data)
        return (blocknum, data)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=C0116,W0613
# This program is dedicated to the public domain under the CC0 license.
import math
import time

import pandas as pd
from web3 import exceptions, _utils

from ..b_send import BSend
from ..b_send.basec import BaseBulk, PrintNetworkName
from ..pharaohs import pharaohs
from ...libeb import MiliDoS


class SkeletonLooper(BaseBulk):
    """
    Bulk manager execution now
    @
    """

    def __init__(self, _core: MiliDoS):
        self._c = _core
        super().__init__()
        self.__n = 0
        self.__t = 0
        self.__failures = 0
        self.wait_pause = False
        PrintNetworkName(_core.network_cfg)

    def failureCounts(self) -> int:
        return self.__failures

    def ActivateWaitPause(self):
        self.wait_pause = True

    def _line_progress(self, notify=None) -> None:
        if notify is None:
            return
        else:
            perc = "{0:.0f}%".format(self.__n / self.__t * 100)
            notify(self.__n, self.__t, perc)

    def _line_error(self, errorNotify=None, info: str = "") -> None:
        if errorNotify is None:
            print(f"======{info}")
        else:
            errorNotify(info)

    def setIterSum(self, t) -> int:
        self.__t = t
        return self.__t

    def Iters(self) -> int:
        return self.__t

    def addFail(self) -> int:
        self.__failures += 1
        return self.__failures

    def addN(self) -> int:
        self.__n += 1
        return self.__n

    def iterN(self) -> int:
        return self.__n

    def resetN(self) -> int:
        self.__n = 0
        return self.__n

    def failure(self, a: str, b: str) -> None:
        pass

    def successTransaction(self, hash: str, name: str) -> None:
        pass


class TestBulkManager(SkeletonLooper):
    """
    Bulk manager execution now
    @
    """

    def __init__(self, dat: list, _core: MiliDoS):
        self.datlist = dat
        super().__init__(_core)
        self._enableContractBatch()

    def prep(self) -> "TestBulkManager":
        self._status_busy = True
        for row in self.datlist:
            address = str(row[0])
            amount = float(row[1])
            enter_digit = int(amount * 10 ** self.decimal)
            if self._is_valid_address(address):
                self._line_read_code(address, amount, enter_digit)
                self.entryAdd(address, enter_digit)
            else:
                self._line_invalid_address(address)
                self.entryErrAdd(address, enter_digit)

        self._batch_preprocess()
        self.PreStatement()

        return self

    def getSENDAddresses(self) -> list:
        return self.list_address

    def getSENDAmountBalances(self) -> list:
        return self.list_amount

    def getPlatformVal(self) -> int:
        """
        since the entry for python function on SAP is required to be int
        """
        return math.ceil(self.total / self.wei)


class ExcelFeature(SkeletonLooper):

    def __init__(self, filepath, _core: MiliDoS):
        super().__init__(_core)
        self.exeFilepath = filepath
        self.useKeyEng()

    def useKeyChinese(self) -> "ExcelFeature":
        self.kAddress = "提现地址"
        self.kAmount = "提现金额"
        return self

    def useKeyEng(self) -> "ExcelFeature":
        self.kAddress = "address"
        self.kAmount = "amount"
        return self


class ExcelBulkManagerContractTunnel(ExcelFeature):
    """
    using contract on making at least XXX transactions in a batch.
    """

    def __init__(self, filepath, m: MiliDoS):
        super().__init__(filepath, m)
        self._enableContractBatch()

    def failure(self, a: str, b: str) -> None:
        """
        custom failure function and recording
        :param a:
        :param b:
        :return:
        """
        if self._file_logger is not None:
            self._file_logger(f"Batch#{self.iterN()} {a} {b} Failed. ❌ ")
        self.addFail()

    def successTransaction(self, hash: str, name: str) -> None:
        """
        custom success function and the recording
        :param hash:
        :param name:
        :return:
        """
        if self._file_logger is not None:
            self._file_logger(f"Batch#{self.iterN()} hash: {hash} 📤 ")

    def executeTokenTransferOnContractBusTg(
            self,
            express_contract: BSend,
            coin_contract: pharaohs,
            notify=None, errorNotify=None) -> None:
        """


        :param express_contract:contract instance
        :param coin_contract: token contract instance
        :param notify: callback function
        :param errorNotify: callback function

        """
        self._status_busy = True
        coin_address = coin_contract.contract_address
        express_address = express_contract.contract_address
        self.setIterSum(self._batches_count)
        self.resetN()

        if not self._batch_contract:
            self._line_error(errorNotify, "⚠️ Batch contract is not activated")
            return

        if not self._is_valid_address(coin_address):
            self._line_error(errorNotify, f"⚠️ ERC20 is not valid {coin_address}")
            return

        for batch in self._batch:
            try:
                batch_size = len(batch[0])
                total_approval = sum(batch[1])
                _address = batch[0]
                _amount = batch[1]

                print(f"====== result batch len: {len(batch[0])}, approving: {total_approval}")
                balance = coin_contract.balance_of(self._c.accountAddr)

                if balance >= total_approval:
                    coin_contract.EnforceTxReceipt(True)
                    coin_contract.approve(express_address, total_approval)
                else:
                    self._line_error(errorNotify, "⚠️ not enough in the balance")
                    return

                print(f"====== start batch transactions")
                express_contract.onSuccssCallback(self.successTransaction)
                express_contract.onFailCallback(self.failure)
                express_contract.EnforceTxReceipt(True).bulk_send_token(
                    coin_address, _address, _amount, 0
                )

                self.addN()
                self._line_progress(notify)

                if batch_size == self.batch_limit and self.wait_pause:
                    print("====== result bulk_send_token, the next batch will start in 30 seconds")
                    time.sleep(30)
                else:
                    print("====== Now the next wave of token send will start immediately")

            except exceptions.CannotHandleRequest:
                self._line_error(errorNotify, "⚠️ request is not handled")
                return
            except exceptions.TimeExhausted:
                self._line_error(errorNotify, "⚠️ the transaction is not on chain after timeout")
                return
            except _utils.threads.Timeout:
                self._line_error(errorNotify, "⚠️ threads timeout")
                return

        self._status_busy = False

    def prep(self) -> "ExcelBulkManagerContractTunnel":
        """
        counting the excel sheet and filter the data
        :return: This is a chained method
        """
        self._status_busy = True
        # df = pd.read_excel(r'C:\Users\Ron\Desktop\Product List.xlsx')
        data = pd.read_excel(self.exeFilepath)
        # data = pd.read_excel(r'C:\Users\Ron\Desktop\Product List.xlsx')
        df = pd.DataFrame(data, columns=[self.kAddress, self.kAmount])
        for index, row in df.iterrows():

            # trim line
            address = str(row[self.kAddress]).translate(str.maketrans('', '', ' \n\t\r'))
            amount = float(row[self.kAmount])
            enter_digit = int(amount * 10 ** self.decimal)
            if self._is_valid_address(address):
                self._line_read_code(row[self.kAddress], amount, enter_digit)
                self.entryAdd(address, enter_digit)
            else:
                self._line_invalid_address(address)
                self.entryErrAdd(address, enter_digit)
                continue

        self._batch_preprocess()
        self.PreStatement()

        return self

    def getSENDAddresses(self) -> list:
        return self.list_address

    def getSENDAmountBalances(self) -> list:
        return self.list_amount

    def getSENDTotal(self) -> int:
        return self.total

    def getPlatformVal(self) -> int:
        return int(self.total / self.wei)


class ExcelBulkManagerClassic(ExcelFeature):
    """
    This is the traditional wallet to wallet transaction transfer using the native method
    """

    def __init__(self, filepath, tron):
        super().__init__(filepath, tron)
        self._in_process_address = None
        self._in_process_amount = 0

    def prep(self) -> "ExcelBulkManagerClassic":
        """
        counting the excel sheet and filter the data
        :return: This is a chained method
        """
        self._status_busy = True
        # df = pd.read_excel(r'C:\Users\Ron\Desktop\Product List.xlsx')
        data = pd.read_excel(self.exeFilepath)
        # data = pd.read_excel(r'C:\Users\Ron\Desktop\Product List.xlsx')
        df = pd.DataFrame(data, columns=[self.kAddress, self.kAmount])
        for index, row in df.iterrows():
            # trim line
            address = str(row[self.kAddress]).translate(str.maketrans('', '', ' \n\t\r'))
            amount = float(row[self.kAmount])
            enter_digit = int(amount * 10 ** self.decimal)
            try:
                if self._is_valid_address(address):
                    self._line_color_code(address, amount, enter_digit)
                    self.entryAdd(address, enter_digit)
                else:
                    self._line_color_invalid_address(address, enter_digit)
                    self.entryErrAdd(address, enter_digit)

            except ValueError as h:
                self._line_color_invalid_address(address, enter_digit, h)
                self.entryErrAdd(address, enter_digit)

        self.PreStatement()

        return self

    def getSENDAddresses(self) -> list:
        return self.list_address

    def getSENDAmountBalances(self) -> list:
        return self.list_amount

    def getSENDTotal(self) -> int:
        return self.total

    def executeTokenDistribution(self, token: pharaohs, notify=None):

        self.resetN()
        self.setIterSum(self.transaction_count)
        self._status_busy = True

        if len(self.list_amount) != len(self.list_address):
            print("error in checking the length of transaction list")
            return

        for address in self.list_address:
            token.transfer(address, self.list_amount[self.iterN()])
            self.addN()
            self._line_progress(notify)

        self._status_busy = False

    def failure(self, a: str, b: str) -> None:
        if self._file_logger is not None:
            self._file_logger(f"#{self.iterN()} {a} {b} Failed. ❌ ")
        self.addFail()

    def successTransaction(self, hash: str, name: str) -> None:
        if self._file_logger is not None:
            self._file_logger(f"#{self.iterN()} {hash} OK, {self._in_process_address} {self._in_process_amount} 📤 ")

    def executeTokenTransferDistributionTg(self, token: pharaohs, notify=None, errorNotify=None) -> None:
        """
         Limitation: https://core.telegram.org/bots/faq#my-bot-is-hitting-limits-how-do-i-avoid-this
         When sending messages inside a particular chat, avoid sending more than one message per second. We may allow short bursts that go over this limit, but eventually you'll begin receiving 429 errors.
         If you're sending bulk notifications to multiple users, the API will not allow more than 30 messages per second or so. Consider spreading out notifications over large intervals of 8—12 hours for best results.
         Also note that your bot will not be able to send more than 20 messages per minute to the same group.

         This is a block function and it will take some time to complete

        :param token: token instance
        :param notify: callback function
        :param errorNotify: callback function
        """
        self.resetN()
        self.setIterSum(len(self.list_address))
        self._status_busy = True

        _timestamp = self.nowSec
        if len(self.list_amount) != self.Iters():
            errorNotify("error in checking the length of transaction list")
            return

        token.onSuccssCallback(self.successTransaction)
        token.onFailCallback(self.failure)

        try:

            for address in self.list_address:
                self._in_process_address = address
                report_amount = self.list_amount[self.iterN()]
                self._in_process_amount = report_amount
                token.transfer(address, report_amount)
                self.addN()
                _dela = self.nowSec
                if notify is not None:
                    if _dela > _timestamp + 5 or self.iterN() >= self.Iters():
                        _timestamp = self.nowSec
                        self._line_progress(notify)

                time.sleep(0.5)

        except ValueError:
            errorNotify("Value error. unknown error")
            return

        self._status_busy = False
